#
# See the file "LICENSE" for the full license governing this code.
#
require 'pathname'
$LOAD_PATH.unshift(Pathname.new(__FILE__).dirname.parent.parent)
$LOAD_PATH.unshift(Pathname.new(__FILE__).dirname.parent.parent.parent.parent + 'easy_type' + 'lib')
require 'puppet/property/boolean'
require 'easy_type'
require 'easy_type/array_property'
require 'easy_type/encryption'
require 'puppet_x/enterprisemodules/oracle/access'
require 'puppet_x/enterprisemodules/oracle/read_only_property'
require 'puppet_x/enterprisemodules/oracle/information'
require 'puppet_x/enterprisemodules/oracle/commands'
require 'puppet_x/enterprisemodules/oracle/ora_tab'
require 'puppet_x/enterprisemodules/oracle/directories'
require 'puppet_x/enterprisemodules/oracle/utilities'

Puppet::Type.newtype(:ora_database) do
  include EasyType
  include EasyType::Encryption
  include Puppet_X::EnterpriseModules::Oracle::Access
  include Puppet_X::EnterpriseModules::Oracle::Directories
  include Puppet_X::EnterpriseModules::Oracle::Commands
  include Puppet_X::EnterpriseModules::Oracle::Information

  desc <<-EOD
    This resource allows you to manage an Oracle Database.

    This type allows you to create a database. In one of it's simplest form:

        ora_database{'oradb':
          ensure          => present,
          oracle_base     => '/opt/oracle',
          oracle_home     => '/opt/oracle/app/11.04',
          control_file    => 'reuse',
        }

    The `ora_database` type uses structured types for some of the parameters. Here is part of an example
    with some of these structured parameters filled in:

        ora_database{'bert':
          logfile_groups => [
              {file_name => 'test1.log', size => '10M'},
              {file_name => 'test2.log', size => '10M'},
            ],
          ...
          default_tablespace => {
            name      => 'USERS',
            datafile  => {
              file_name  => 'users.dbs',
              size       => '1G',
              reuse      =>  true,
            },
            extent_management => {
              type          => 'local',
              autoallocate  => true,
            }
          },
          ...
          datafiles       => [
            {file_name   => 'file1.dbs', size => '1G', reuse => true},
            {file_name   => 'file2.dbs', size => '1G', reuse => true},
          ],
          ...
          default_temporary_tablespace => {
            name      => 'TEMP',
            type      => 'bigfile',
            tempfile  => {
              file_name  => 'tmp.dbs',
              size       => '1G',
              reuse      =>  true,
              autoextend => {
                next    => '10K',
                maxsize => 'unlimited',
              }
            },
            extent_management => {
              type          => 'local',
              uniform_size  => '1G',
            },
          },
          ....
          undo_tablespace   => {
            name      => 'UNDOTBS',
            type      => 'bigfile',
            datafile  => {
              file_name  => 'undo.dbs',
              size       => '1G',
              reuse      =>  true,
            }
          },
          ....
          sysaux_datafiles => [
            {file_name   => 'sysaux1.dbs', size => '1G', reuse => true},
            {file_name   => 'sysaux2.dbs', size => '1G', reuse => true},
          ]

  EOD

  to_get_raw_resources do
    #
    # If Puppet::Type::Ora_setting is already defined, it means the type is already loaded. If we 
    # require the file again, Puppet issues an error. 
    #
    if defined?(Puppet::Type::Ora_setting).nil?
      require 'puppet/type/ora_setting'
    end
    #
    # Check for all available SID's. Not all SID's need to be running
    #
    available_databases.collect do |db|
      contained_by_value = configuration[db]['contained_by']
      if running_db?(db) || local_pdb?(db)
        base_command = <<-EOD
column name format a30
column log_mode format a13
column cdb format a3
column force_logging format a13
column flashback_on format a14
EOD
        command = if local_pdb?(db)
                    "select \'#{db}\' as name,
                            'NA' as log_mode,
                            null as force_logging,
                            null as flashback_on,
                            'running' state
                       from gv$pdbs gvp
                      where upper(gvp.name) = \'#{db}\'"
                  else
                    "select case (select value from v$parameter where name = 'cluster_database')
                              when 'TRUE' then name
                              else \'#{db}\'
                            end as name,
                            gvd.log_mode,
                            gvd.force_logging,
                            gvd.flashback_on,
                            decode(gvd.open_mode,'READ WRITE','running','MOUNTED','mounted',gvd.open_mode) state
                       from gv$database gvd,
                            gv$instance gvi
                      where gvd.inst_id = gvi.inst_id
                        and upper(gvi.instance_name) = \'#{db}\'"
                  end
        full_command = base_command + command
        entry = sql(full_command, :sid => db).first
        entry['CONTAINED_BY'] = contained_by_value
        entry
      else
        EasyType::Helpers::InstancesResults['SID', db, 'NAME', db, 'LOG_MODE', nil, 'STATE', 'stopped', 'CDB', 'disabled', 'FORCE_LOGGING', 'disabled', 'FLASHBACK_ON', 'NO', 'CONTAINED_BY', contained_by_value]
      end
    end
  end

  ensurable

  on_create do
    @dbname = cluster? ? instance_name : name
    overwrite_check
    add_settings_entry
    if contained_by.to_s != ''
      fail 'archivelog cannot be specified for pluggable databases' if ['enabled','disabled'].include? archivelog
      self[:sys_password] = configuration_value_for(contained_by, 'password')
      self[:oracle_home] = configuration_value_for(contained_by, 'oracle_home')
      sql template('ora_config/ora_database/create_pdb.sql.erb', binding), :sid => contained_by
    else
      begin
        add_oratab_entry
        create_directories
        create_init_ora_file
        create_ora_pwd_file

        create_stage_1
        create_stage_2
        execute_stage_1
        execute_stage_2

      rescue => e
        remove_directories unless contained_by
        remove_services if Facter.value(:kernel) == 'windows'
        remove_oratab_entry
        remove_settings_entry
        fail "Error creating database #{name}, #{e.message}. Rolling back creation of database"
      end

      begin
        if cluster?
          register_database
        end
      rescue => e
        fail "Error registering database #{name} as cluster resource, #{e.message}"
      end

      begin
        if cluster?
          add_instances
        end
      rescue => e
        fail "Error registering instance(s) for database #{name} as cluster resource, #{e.message}"
      end

      begin
        if cluster?
          start_database
        end
      rescue => e
        fail "Error starting database #{name} as cluster resource, #{e.message}"
      end

  end
    nil
  end

  on_modify do
    nil # Allow properties to do their stuff
  end

  on_destroy do
    if running_database_sids.include?(name)
      oracle_base ||= diagnostic_dest(name)
    end
    if provider.contained_by.empty?
      @dbname = cluster? ? instance_name : name
      if cluster?
        remove_instance_registrations
        remove_database_registration
      end
      sql template('ora_config/ora_database/destroy.sql.erb', binding), :sid => name, :daemonized => false
      if Facter.value(:kernel) == 'windows'
        rm_instance = remove_services
      end
      remove_directories
      remove_oratab_entry
    else
      sql template('ora_config/ora_database/destroy_pdb.sql.erb', binding), :sid => provider.contained_by, :daemonized => false
    end
    remove_settings_entry
    nil
  end

  parameter :name
  parameter :overwrite
  parameter :system_password
  parameter :sys_password
  parameter :init_ora_content
  parameter :timeout
  parameter :control_file

  property  :character_set
  property  :national_character_set
  property  :logfile_groups
  property  :maxdatafiles
  property  :maxinstances
  property  :maxlogfiles
  property  :maxlogmembers
  property  :maxloghistory
  property  :timezone
  #
  # For managing the ora_settings
  #
  parameter :default
  parameter :daemonized

  parameter :tablespace_type
  parameter :logfile
  parameter :extent_management
  parameter :oracle_home
  parameter :oracle_base
  parameter :oracle_user
  parameter :oracle_user_password
  parameter :install_group
  parameter :default_tablespace
  parameter :datafiles
  parameter :default_temporary_tablespace
  parameter :undo_tablespace
  parameter :sysaux_datafiles
  parameter :spfile_location
  parameter :config_scripts
  parameter :options
  #
  # Stuff for container databases
  #
  property  :container_database
  parameter :file_name_convert
  #
  # Stuff for pluggable databases
  #
  property  :contained_by
  parameter :pdb_admin_user
  parameter :pdb_admin_password
  # When defining a RAC database, these become valuable
  #
  parameter :instances
  parameter :scan_name
  parameter :scan_port
  #
  # Properties
  #
  property :state
  property :force_logging
  property :autostart
  group(:arch_group) do
    property :archivelog
    property :flashback
  end

  private

  def self.available_databases
    sids = registered_sids
    sids.select { |sid| database_sid?(sid) || local_pdb?(sid) }
  end

  #
  # Because creating a database is a potential destructive action, we add some extra checks
  #
  def overwrite_check
    return if self[:overwrite] == true
    fail "#{oracle_home}/dbs/orapw#{@dbname} already exists. Not overwriting" if File.exists?("#{oracle_home}/dbs/orapw#{@dbname}")
    fail "#{init_ora_file} already exists. Not overwriting." if File.exists?(init_ora_file)
    fail "#{oracle_base}/admin/#{name} or other sid related directories already exists. Not overwriting." if oracle_directories_exists?
  end

  def add_settings_entry
    contained_by = self[:contained_by] # temporary shadow parameter
    if contained_by.to_s != ''
      pluggable        = true
      container_domain = db_domain(contained_by)
      domain_name      = container_domain.to_s.empty? ? '' : ".#{container_domain}"
      oracle_home      = configuration_value_for(contained_by, 'oracle_home')
    else
      oracle_home  = self[:oracle_home]
      pluggable    = false
      domain_name  = ''
      contained_by = ''
    end
    #
    # We only allow username and password from remote databases
    #
    connect_string = contained_by.empty? ? '' : "//#{Facter.value('hostname')}/#{name}#{domain_name}"
    cdb = container_database.to_s == 'enabled' ? true : false
    default = self[:default] || num_default_database_sids <= 1 

    if Facter.value(:kernel) == 'windows'
      entry = Puppet::Type::Ora_setting.new(:name           => @dbname,
                                            :os_user        => oracle_user,
                                            :os_user_pwd    => oracle_user_password,
                                            :oracle_home    => oracle_home,
                                            :default        => default,
                                            :connect_string => connect_string,
                                            :daemonized     => self[:daemonized],
                                            :pluggable      => pluggable,
                                            :cdb            => cdb,
                                            :contained_by   => contained_by)
    else
      entry = Puppet::Type::Ora_setting.new(:name           => @dbname,
                                            :os_user        => oracle_user,
                                            :oracle_home    => oracle_home,
                                            :default        => default,
                                            :connect_string => connect_string,
                                            :daemonized     => self[:daemonized],
                                            :pluggable      => pluggable,
                                            :cdb            => cdb,
                                            :contained_by   => contained_by)

    end
    entry.add
  end

  def remove_settings_entry
    entry = Puppet::Type::Ora_setting.new(:name => name)
    entry.remove
  end

  def create_init_ora_file
    File.open(init_ora_file, 'w') do |file|
      file.write(init_ora_content)
      if cluster?
        write_cluster_parameters(file)
      else
        file.write("#\n")
        file.write("# Parameters inserted by Puppet ora_database\n")
        file.write("#\n")
      end
    end
    owned_by_oracle(init_ora_file)
    Puppet.debug "ora_database.rb: File #{init_ora_file} created with content"
  end

  def write_cluster_parameters(file)
    instance_names = instances.keys.sort # sort the keys for ruby 1.8.7 Hash ordering
    instance_names.each_index do |index|
      instance = instance_names[index]
      instance_no = index + 1
      file.write("#\n")
      file.write("# Parameters inserted by Puppet ora_database (RAC)\n")
      file.write("#\n")
      file.write("#{instance}.instance_number=#{instance_no}\n")
      file.write("#{instance}.thread=#{instance_no}\n")
      file.write("#{instance}.undo_tablespace=UNDOTBS#{instance_no}\n")
    end
  end

  def add_oratab_entry
    unless Puppet::Util::Platform.windows?
      oratab = Puppet_X::EnterpriseModules::Oracle::OraTab.new
      oratab.ensure_entry(@dbname, oracle_home, autostart)
    end
  end

  def remove_oratab_entry
    unless Puppet::Util::Platform.windows?
      oratab = Puppet_X::EnterpriseModules::Oracle::OraTab.new
      oratab.remove_entry(@dbname, oracle_home, autostart)
    end
  end

  def remove_services
    Puppet::Util::Execution.execute("#{oracle_home}\\bin\\oradim.exe -delete -sid #{@dbname}", opts)
  end

  def create_ora_pwd_file
    if Puppet::Util::Platform.windows?
      orapwd("file=#{oracle_home}\\database\\PWD#{@dbname}.ora force=y password='#{sys_password}' entries=20", :sid => @dbname)
    else
      # TODO: if cluster/oracle restart or using ASM store pwfile in ASM
      orapwd("file=#{oracle_home}/dbs/orapw#{@dbname} force=y password='#{sys_password}' entries=20", :sid => @dbname)
    end
  end

  def create_stage_1
    osfamily = Facter.value('osfamily')
    content = template('ora_config/ora_database/create.sql.erb', binding)
    path = "#{admin_scripts_path}/create.sql"
    File.open(path, 'w') { |f| f.write(content) }
    owned_by_oracle(path) unless Puppet::Util::Platform.windows?
  end

  def create_stage_2
    with_config_scripts do |script, content|
      path = "#{admin_scripts_path}/#{script}.sql"
      File.open(path, 'w') { |f| f.write(content) }
      owned_by_oracle(path) unless Puppet::Util::Platform.windows?
    end
  end

  def start_database
    srvctl("start database -d #{name}", :sid => @dbname)
  end

  def register_database
    command = if spfile_location
                "add database -d #{name} -o #{oracle_home} -n #{name} -m #{name} -p #{spfile_location}/spfile#{name}.ora "
              else
                "add database -d #{name} -o #{oracle_home} -n #{name} -m #{name} "
              end
    srvctl(command, :sid => @dbname)
  end

  def add_instances
    instances.each do |instance, node|
      if node == hostname
        srvctl("add instance -d #{name} -i #{instance} -n #{node}", :sid => @dbname)
      end
    end
  end

  def disable_database
    srvctl("disable database -d #{name}", :sid => @dbname)
  end

  def remove_instance_registrations
    instances.each do |instance, _node|
      srvctl("remove instance -d #{name} -i #{instance}", :sid => @dbname)
    end
  end

  def remove_database_registration
    srvctl("remove database -d #{name}", :sid => @dbname)
  end

  def execute_stage_1
    #
    # The create script generates some error's. SO that is why we ignore the errors
    #
    if Puppet::Util::Platform.windows?
      create_instance = Puppet::Util::Execution.execute("#{oracle_home}\\bin\\oradim.exe -new -sid #{@dbname} -startmode manual -spfile", opts)
      edit_instance = Puppet::Util::Execution.execute("#{oracle_home}\\bin\\oradim.exe -edit -sid #{@dbname} -startmode auto -srvcstart system", opts)
    end
    sql("@#{admin_scripts_path}/create", :sid => @dbname, :timeout => 0, :failonsqlfail => false, :daemonized => false)
  end

  def execute_stage_2
    with_config_scripts do |script, _|
      sql("@#{admin_scripts_path}/#{script}", :sid => @dbname, :timeout => 0, :daemonized => false)
    end
  end

  def instance_name
    instances.keys.sort.first
  end

  def cluster?
    instances.count > 0
  end

  def hostname
    Facter.value('hostname')
  end

  def opts
    { :failonfail => true, :combine => true }
  end

  def init_ora_file
    if Puppet::Util::Platform.windows?
      "#{oracle_home}\\database\\init#{@dbname}.ora"
    else
      "#{oracle_home}/dbs/init#{@dbname}.ora"
    end
  end

  def admin_scripts_path
    "#{oracle_base}/admin/#{name}/scripts"
  end

  def with_config_scripts
    config_scripts.each do |entry|
      script = entry.keys.first
      content = entry[script]
      yield(script, content)
    end
  end

  def asm_path?(path)
    path.include?('+') ? true : false
  end
end
