#
# See the file "LICENSE" for the full license governing this code.
#
newproperty(:status) do
  include EasyType
  include Puppet_X::EnterpriseModules::Oracle::Access
  include Puppet_X::EnterpriseModules::Oracle::Commands
  include Puppet_X::EnterpriseModules::Oracle::Information

  desc <<-EOD
    The state of the service. It can be either running or stopped. On a primary cluster
    instance the default is `running`. This means it will automaticaly start the service. When the
    instance is a standby instance, and you do not specify an expliciet value, Puppet will leave
    the state of the instance as it is.
  EOD

  newvalues(:running, :stopped)
  defaultto {
    begin
      return :running if primary?
    rescue Puppet::ExecutionFailure
      Puppet.debug "#{path}: Oracle not running. No way to determine best default. Choosing no default"
      return nil
    rescue RuntimeError
      Puppet.debug "#{path}: SID not defined yet. No way to determine best default. Choosing no default"
      return nil
    end
  }
  data_type("Optional[Enum['running', 'stopped']]")

  to_translate_to_resource do |raw_resource|
    #
    # When we are looking at a CDB service, we use the SID of the container instead of the
    # sid of the pdb.
    #
    sid = raw_resource['container'] || raw_resource['SID']
    if cluster?(sid)
      name = raw_resource['NAME'].upcase
      cluster_service_running_on(name, sid).empty? ? :stopped : :running
    end
  end

  after_create do
    return true unless cluster?(resource.sid)
    after_modify
  end

  after_modify do
    case resource[:status].to_s
    when 'stopped'
      stop_service if needs_stopping?
    when 'running'
      stop_service if needs_stopping?
      start_service
    else
      fail 'internal error. Invalid status found'
    end
  end

  def sid
    resource.sid
  end

  def insync?(_is)
    return true unless cluster?
    if should.to_s == 'running'
      all_started?
    else
      all_stopped?
    end
  end

  def change_to_s(current, should)
    return "Starting on #{(needs_starting).join(',')}" if needs_starting?
    return "Stopping on #{(needs_stopping).join(',')}" if needs_stopping?
    fail "Internal Error: Invalid state change."
  end

  def start_service
    srvctl("start service -d #{db_for(sid)} -s #{resource[:service_name]}", :sid => sid_for(sid))
  end

  def stop_service
    srvctl("stop service -d #{db_for(sid)} -s #{resource[:service_name]}", :sid => sid_for(sid))
  end

  def all_started?
    cluster_service_running_on.sort == preferred_instances.sort - unknown_instances.sort
  end

  def all_stopped?
    cluster_service_stopped_on.sort == preferred_instances.sort - unknown_instances.sort
  end

  def cluster_service_running_on
    self.class.cluster_service_running_on(resource[:service_name], resource[:sid])
  end

  def cluster_service_stopped_on
    preferred_instances - cluster_service_running_on
  end

  def needs_starting?
    needs_starting.any?
  end

  def needs_stopping?
    needs_stopping.any?
  end

  def needs_starting
    preferred_instances.sort - cluster_service_running_on.sort
  end

  def needs_stopping
    cluster_service_running_on.sort
  end

  def self.cluster_service_running_on(name, local_sid = sid)
    running_on = status_string(local_sid).scan(/Service #{name} is running on instance\(s\) (.*)$/i).first
    if running_on
      running_on.first.split(',')
    else
      []
    end
  end

  def self.status_string(sid)
    if pdb?(sid)
      pdb = "-pdb #{sid}"
    end
    @status_string ||= srvctl("status service -d #{db_for(sid)} #{pdb}", :sid => sid_for(sid))
  end

  def preferred_instances
    resource[:preferred_instances] ? resource[:preferred_instances] : cluster_instances
  end

  def unknown_instances
    preferred_instances - cluster_instances
  end
end
