#!/opt/puppetlabs/puppet/bin/ruby
#
# See the file "LICENSE" for the full license governing this code.
#
require 'json'
require 'puppet'
USER_SKIP_LIST = [
  'SYS',
  'SYSTEM',
  'XDB',
  'GSMCATUSER',
  'GSMUSER',
  'AUDSYS',
  'DIP',
  'ORACLE_OCM',
  'ANONYMOUS',
  'SYSBACKUP',
  'SYSRAC',
  'SYSKM',
  'SYS$UMF',
  'SYSDG',
  'DBSNMP',
  'APPQOSSYS',
  'DBSFWUSER',
  'GGSYS',
  'XS$NULL',
  'OUTLN',
  'GSMADMIN_INTERNAL',
  'REMOTE_SCHEDULER_AGENT',
  'GSMROOTUSER',
]

install_dir = ENV['PT__installdir']
Puppet[:codedir] = install_dir
Puppet[:vardir]  = '/tmp'

if install_dir
  #
  # Use loadpath specfied in the params
  #
  $LOAD_PATH << "#{install_dir}/easy_type/lib"
  $LOAD_PATH << "#{install_dir}/ora_config/lib"
else
  begin
    Puppet.initialize_settings
  rescue Puppet::DevError
    # When run in Rspec, this is called multiple times
  end
end
require 'easy_type'

begin
  params           = JSON.parse(STDIN.read)
  sid              = params.fetch('sid')
  oracle_home      = params.fetch('oracle_home')
  cdb              = params.fetch('cdb')
  os_user          = params.fetch('os_user')
  pluggable        = params.fetch('pluggable')
  contained_by     = params.fetch('contained_by')
  system_password  = params.fetch('system_password')
  sys_password     = params.fetch('sys_password')
  os_user_password = params.fetch('os_user_password')

  init_param_filter = ->(e) do
    return nil if e[e.keys.first][:scope] != 'SPFILE'
    return nil if e[e.keys.first][:sid] != sid
    e
  end 

  user_filter       = ->(e) do
    return nil if USER_SKIP_LIST.any? { |skip| e.keys.first =~ /#{Regexp.escape(skip)}@.*$/ }
    e[e.keys.first].delete(:password)
    e
  end

  tablespace_filter       = ->(e) do
    # A OMF managed tablespace file returns an empty key. We need to remove it
    # here
    e[e.keys.first].delete(:'')
    e
  end

  database_filter   = ->(e) do
    e[e.keys.first][:sys_password]         = sys_password
    e[e.keys.first][:system_password]      = system_password
    e[e.keys.first][:oracle_user_password] = os_user_password
    #
    # We want ora_database to use the provided home of the target, not of the source
    #
    e[e.keys.first][:oracle_user] = 'oracle'                                            # On the new systems we use oracle
    e[e.keys.first][:oracle_base] = "%{lookup('ora_profile::database::oracle_base')}"   # On the new systems we use our own oracle_base
    e[e.keys.first][:oracle_home] = "%{lookup('ora_profile::database::oracle_home')}"   # On the new systems we use our own oracle_home
    e
  end 

  resources = {
    # :ora_autotask        => { :key => 'ora_profile::database::extracted_autotasks' },
    # :ora_awr_settings    => { :key => 'ora_profile::database::extracted_awr_settings' },
    # :ora_feature_usage   => { :key => 'ora_profile::database::extracted_' },
    # :ora_function        => { :key => 'ora_profile::database::extracted_' },
    # :ora_home_option     => { :key => 'ora_profile::database::extracted_' },
    # :ora_package         => { :key => 'ora_profile::database::extracted_packages' },
    # :ora_procedure       => { :key => 'ora_profile::database::extracted_procedures' },
    # :ora_thread          => { :key => 'ora_profile::database::extracted_threads' },
    # :ora_trigger         => { :key => 'ora_profile::database::extracted_triggers::list' },

    :ora_database        => { :key => 'ora_profile::database::extracted_database_definition::list', :filter => database_filter },
    :ora_directory       => { :key => 'ora_profile::database::extracted_directories::list' },
    :ora_init_param      => { :key => 'ora_profile::database::extracted_init_params::list', :filter => init_param_filter },
    :ora_object_audit    => { :key => 'ora_profile::database::extracted_object_autits::list' },
    :ora_object_grant    => { :key => 'ora_profile::database::extracted_object_grants::list' },
    :ora_profile         => { :key => 'ora_profile::database::extracted_profiles::list' },
    :ora_role            => { :key => 'ora_profile::database::extracted_roles::list' },
    :ora_scheduler_job   => { :key => 'ora_profile::database::extracted_scheduler_jobs::::list' },
    :ora_service         => { :key => 'ora_profile::database::extracted_services::list' },
    :ora_statement_audit => { :key => 'ora_profile::database::extracted_statement_audits::list' },
    :ora_synonym         => { :key => 'ora_profile::database::extracted_synoniems::list' },
    :ora_tablespace      => { :key => 'ora_profile::database::extracted_tablespaces::list', :filter => tablespace_filter },
    :ora_user            => { :key => 'ora_profile::database::extracted_users::list', :filter => user_filter },
  }
  settings_file = '/etc/ora_setting.yaml'
  
  configuration = { 
    sid=> { 
      'oracle_home'    => oracle_home, 
      'cdb'            => cdb, 
      'os_user'        => os_user, 
      'pluggable'      => pluggable, 
      'contained_by'   => contained_by, 
      'default'        => true,
      'connect_string' => '',
      'user'           => 'sys',
      'syspriv'        => 'sysdba',
      'daemonized'     => false
    }
  }
  File.open(settings_file, "w") { |file| file.write(configuration.to_yaml) }
  data = EasyType::ExtractTask.new(resources).execute
  File.delete(settings_file)
  puts({ :resources => data }.to_json)
rescue Puppet::Error => e
  puts({ :status => 'failure', :error => e.message }.to_json)
  exit 1
end